/*  File: utils.c
 *  Author: Ed Griffiths (edgrif@sanger.ac.uk)
 *  Copyright (c) J Thierry-Mieg and R Durbin, 1999
 * -------------------------------------------------------------------
 * Acedb is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * or see the on-line version at http://www.gnu.org/copyleft/gpl.txt
 * -------------------------------------------------------------------
 * This file is part of the ACEDB genome database package, written by
 * 	Richard Durbin (Sanger Centre, UK) rd@sanger.ac.uk, and
 *	Jean Thierry-Mieg (CRBM du CNRS, France) mieg@kaa.crbm.cnrs-mop.fr
 *
 * Description: Contains utility functions that are "one-offs" and not
 *              really part of any other package.
 * Exported functions: getLogin
 * HISTORY:
 * Last edited: Mar  1 14:05 2001 (edgrif)
 * * Mar 22 14:42 1999 (edgrif): Added getSystemName().
 * * Jan 27 09:20 1999 (edgrif): Add inokedebugger/regExpMatch from messubs.
 * * Jan 25 16:05 1999 (edgrif): Add getLogin from session.c
 * Created: Thu Jan 21 15:46:49 1999 (edgrif)
 * CVS info:   $Id: utils.c,v 1.26 2001/03/01 16:26:39 edgrif Exp $
 *-------------------------------------------------------------------
 */

#include "regular.h"

#ifdef SGI
#include "math.h"					    /* work around SGI library bug */
#endif

#ifdef IBM
/* I can't believe this is still required for IBM machines...sigh..          */

struct passwd {
  char    *pw_name;
  char    *pw_passwd;
  long   pw_uid;
  long   pw_gid;
  char    *pw_gecos;
  char    *pw_dir;
  char    *pw_shell;
} ;
extern struct passwd *getpwuid (long uid) ;
extern struct passwd *getpwnam (const char *name) ;

#else  /* !IBM */

#include <pwd.h>

#endif /* !IBM */


/* This is all pretty unsafe at the moment, it relies on the user having     */
/* called getLogin to initialise the below two globals before using them..   */
/* These globals are also separately set by session.c using the system       */
/* calls getuid() and geteuid(), this is not optimal.                        */
/*                                                                           */
uid_t ruid = -1, euid = -1;

#ifdef ED_G_NEVER_INCLUDE_THIS_CODE
/* Currently unused, there should be a unified interface to uids but this    */
/* requires more understanding of how they are used in the code before       */
/* implementing.                                                             */
uid_t utGetRuid(void)
{
  return ruid ;
}

void utSetRuid(uid_t newRuid)
{
  ruid = newRuid ;
}

uid_t utGetEuid(void)
{
  return euid ;
}

void utSetEuid(uid_t newEuid)
{
  euid = newEuid ;
}
#endif /* ED_G_NEVER_INCLUDE_THIS_CODE */



/*************************************************************/
/****************** getLogin() *******************************/

/**** getLogin() for real/effective user names 
***** original by 03.05.1992 kocab DKFZ  
*****/

char *getLogin (BOOL isReal)
     /* general public function */
{
  /* isReal for real or effective can't fail */

#if defined(MACINTOSH)

  return "acedb" ;
  
#elif defined (WIN32)

  char *name = NULL;
  if ((name = getenv("USERNAME")) != NULL )
    return name ;   /* Windows NT has usernames */
  else
    return "acedb" ;

#else  /* all UNIX */
  static char *rname = 0 ;
  static char *ename = 0 ;

/* RD 980417: changed so getlogin() was last resort.  It can return
   "root" inappropriately where getpwuid(ruid)->pw_name returns the
   right answer (ddts problem SANgc02099)
*/

  if (!ename)
    {
      if (ruid == -1)
	{ ruid = getuid() ;
	euid = geteuid () ;
	}
      
      if (!rname)
	if (ruid)
	  if (getpwuid(ruid))
	    rname = strnew(getpwuid(ruid)->pw_name, 0) ;
      if (!rname)
	if (getlogin())
	  rname = strnew(getlogin(), 0) ;
      if (!rname)
	rname = "getLoginFailed" ;
      
      if (!ename)
	if (getpwuid(euid))
	  ename = strnew(getpwuid(euid)->pw_name, 0) ;
      if (!ename)
	ename = rname ;
      /* mieg: jan5 99, new problem
	 on my dec alpha, if i run as a daemon, getLogin fails totally
	 printf ("rname=%s ename =%s",rname?rname:"null",ename ? ename:"null") ; 
	 */
    }

  return isReal ? rname : ename ;
#endif /* UNIX */

} /* getLogin */


/* Get the system name using the POSIX uname call.                           */
char *getSystemName(void)
{
  static char *sysname = NULL ;
  
  if (sysname == NULL)
    {
      struct utsname sys_details ;
      
      if (uname(&sys_details) != 0)
	{
	  /* POSIX failed... */
	  sysname = messalloc(101);
	  if (gethostname(sysname, 100) == -1)
	    messcrash("cannot retrieve system name.") ;
      }
      else
	sysname = strnew(sys_details.nodename, 0) ;
    }
  
  return sysname ;
}

char *getUserHomeDir (char *user_name, STORE_HANDLE h)
{
  struct passwd *pwd;
  char *homedir = NULL;

  if ((pwd = getpwnam (user_name)))
    homedir = strnew (pwd->pw_dir, h);
  else
    messerror ("Unknown user: %s", user_name);

  return homedir;
} /* getUserHomeDir */

/*****************************/

BOOL getCmdLineOption (int *argcp, char **argv,
		       char *arg_name, char **arg_val)
     /* call with (&argc, argv, "-option", &string)
      *   for options with value (e.g. -display <name>)
      * or with   (&argc, argv, "-option", NULL)
      *   for options that are simple switches (e.g. -help)
      * RETURNS:
      *   TRUE if option was found (will set *arg_val if non-NULL)
      * SIDE-EFFECT:
      *   The option (and possibly its value) will be removed from
      *   the aguments (argc,argv). Call only once per option.
      */
{
  int i, num;
  BOOL isFound = FALSE;

  for (i = 1; i < *argcp; i++)
    {
      if (strcmp(argv[i], arg_name) == 0)
	{
	  if (arg_val)
	    {
	      if ((*argcp - i) < 2)
		messExit ("No argument specified for %s option", 
			  arg_name);

	      *arg_val = strnew(argv[i+1], 0);
	      num = 2;
	    }
	  else
	    num = 1;

	  /* clear argument(s) from the list */
	  for (i += num; i < *argcp; i++)
	    argv[i - num] = argv[i];
	  argv[*argcp - num] = 0;
	  (*argcp) -= num;

	  isFound = TRUE;
	  break;
        }
    }
  return isFound;
} /* getCmdLineOption */


/*****************************/

/* Check to see if user supplied the "-sleep secs" option, this is useful    */
/* for debugging when the acedb program is called from within a script. It   */
/* means that you can make the acedb program sleep until you can attach a    */
/* debugger to see what its doing.                                           */
/*                                                                           */
/* This routine is not perfect but its only intended for debugging use.      */
/*                                                                           */
void checkCmdLineForSleep(int *argcp, char **argv)
{
  char *secs_str = NULL ;
  int secs ;

  if (getCmdLineOption(argcp, argv, "-sleep", &secs_str))
    {
      if (secs_str != NULL)
	secs = atoi(secs_str) ;

      if (secs == 0)					    /* atoi returns 0 on error. */
	secs = 60 ;

      sleep(secs) ;
    }

  return ;
}


/************************************************************/

void invokeDebugger (void) 
/* put "break invokeDebugger" in your favourite debugger init file */
{
  static BOOL reentrant = FALSE ;

  if (!reentrant)
    { reentrant = TRUE ;
      messalloccheck() ;
      reentrant = FALSE ;
    }
}

/*****************************/
/* match to reg expression 

   returns 0 if not found
           1 + pos of first sigificant match (i.e. not a *) if found
*/

int regExpMatch (char *cp,char *tp)
{
  char *c=cp, *t=tp;
  char *ts=0, *cs=0, *s = 0 ;
  int star=0;

  while (TRUE)
    switch(*t)
      {
      case '\0':
 	if(!*c)
	  return  ( s ? 1 + (s - cp) : 1) ;
	if (!star)
	  return 0 ;
        /* else not success yet go back in template */
	t=ts; c=cs+1;
	if(ts == tp) s = 0 ;
	break ;
      case '?' :
	if (!*c)
	  return 0 ;
	if(!s) s = c ;
        t++ ;  c++ ;
        break;
      case '*' :
        ts=t;
        while( *t == '?' || *t == '*')
          t++;
        if (!*t)
          return s ? 1 + (s-cp) : 1 ;
        while (freeupper(*c) != freeupper(*t))
          if(*c)
            c++;
          else
            return 0 ;
        star=1;
        cs=c;
	if(!s) s = c ;
        break;
#ifdef CRAZY_TEST
      case 'A' :
	if (!*c || (*c < 'A' || *c > 'Z'))
	  return 0 ;
	if(!s) s = c ;
        t++ ;  c++ ;
        break;
#endif
      default  :
        if (freeupper(*t++) != freeupper(*c++))
          { if(!star)
              return 0 ;
            t=ts; c=cs+1;
	    if(ts == tp) s = 0 ;
          }
	else
	  if(!s) s = c - 1 ;
        break;
      }
}


/* This SGI workaround make the comiler crash! I've disabled it for now,
   pending determination if it is still needed, or fixes a long-lost library
   bug. - srk */
#if 0

/* work around SGI library bug */
#ifdef  SGI 
double log10 (double x) { return log(x) / 2.3025851 ; }
#endif

#endif


/*************************************************************/

/* calling utilsInit will initialise all the statics makeing
   later calls to getLogin and getSystemName effectively threadsafe
   */

void utilsInit (void)
{
  static int n = 0 ;

  if (n)
    messcrash ("utilsInit is not thread safe, it should be called only once") ;
  n = 1 ;
  getSystemName () ;
  getLogin (TRUE) ;  
}

/****************** little arithmetic routines ************************/

int utArrondi (float x)
     /*   1.3 ->  1 ;  1.7 ->  2
      *  -1.3 -> -1 ; -1.7 -> -2 */
{
  if (x >= 0)
    return (int) (x + 0.5) ;
  else
    return (int) (x - 0.5) ;
}

/********************************************/

int utMainPart (float p)
 /* returns 1, 2, 5, 10, 20, 50, 100 etc 
  * in absolute value the returned number is smaller than p
  *   19 -> 10  ; -19 -> -10  */
{
  register int i = 1, sign = 1;
  
  if (!p)
    return 0 ;
  if (p < 0) 
    { sign  = -1 ; p = -p ; }
  while(i <= p + .00001) 
    i = 10 * i;
  i /= 10 ;

  if (2 * i > p)
    return i * sign;
  if(5 * i > p)
    return 2 * i * sign;
  return
    5*i*sign;
}

/********************************************/

int utMainRoundPart (float p)
 /* returns 1, 2, 5, 10, 20, 50, 100 etc 
  * but the returned number may be a bit bigger than p
  *   19 -> 20  ; -19 -> -20  */
{
  register int i = 1, sign = 1;

  if (!p)
    return 0 ;
  if (p < 0) 
    {sign = -1 ; p = -p ; } /* RD: ambiguous without spaces on SGI */
  while (i < p + .1) 
    i = 10 * i;
  if (4 * i < 5 * p) 
    return i * sign;
  i /= 2;
  if (4 * i < 5 * p) 
    return i * sign;
  i =  (2 * i) / 5;
  if (4 * i < 5 * p)
    return i * sign;
  i = i/2; 
  return i * sign;
}

/*********************************************************/
double utDoubleMainPart (double p)
     /* returns 1 , 2 , 5 ,10 , 20, 50 ,100 etc */
{
  register double i = 1, sign = 1;

  if (!p)
    return 0.;
  if(p < 0) 
    {sign = -1; p = -p;}
  i = (double) exp(log((double)10.0) * (double)(1 + (int)(log10((double)p)))) ;
  if (4 * i < 5 * p) 
    return i * sign;
  i /= 2;
  if (4 * i < 5 * p)
    return i * sign;
  i = (2 * i) / 5;
  if (4 * i < 5 * p)
    return i*sign;
  i = i/2;
  return i*sign;
}



/************************************************************/
/********************* eof ***********************************/
