/*  file: smap.c
 *  Author: Richard Durbin (rd@sanger.ac.uk)
 *  Copyright (c) J Thierry-Mieg and R Durbin, 1999
 * -------------------------------------------------------------------
 * Acedb is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * or see the on-line version at http://www.gnu.org/copyleft/gpl.txt
 * -------------------------------------------------------------------
 * This file is part of the ACEDB genome database package, written by
 * 	Richard Durbin (Sanger Centre, UK) rd@sanger.ac.uk, and
 *	Jean Thierry-Mieg (CRBM du CNRS, France) mieg@kaa.crbm.cnrs-mop.fr
 *
 * Description: 
 * Exported functions: See smap.h
 * HISTORY:
 * Last edited: Jan 29 22:41 2001 (rd)
 * * Aug 26 20:20 (rd): added support for Source_exons
 * * Aug 23 21:15 (rd): added sMapDNA and changed mismatch semantics
 * * Aug 23 19:50 (rd): added bsIndexFind calls for efficiency
 			these are VERY IMPORTANT
 * * Aug 23 19:34 (rd): added sMapKeySet()
 * Created: Wed Jul 28 22:11:20 1999 (rd)
 * CVS info:   $Id: smap.c,v 1.5 2001/08/29 09:29:14 rd Exp $
 *-------------------------------------------------------------------
 */

#include "wh/regular.h"
#include "wh/array.h"
#include "wh/bs.h"
#include "wh/smap.h"
#include "whooks/systags.h"
#include "wh/dna.h"
#include "wh/bindex.h"

/* The SMap package supports building a sequence coordinate system out
   of multiple objects according to the following model:

in parent object
  S_Child <tag2> <key> XREF <ctag2> UNIQUE Int UNIQUE Int #SMap_info
  	     // Ints are start, end in parent

#SMap_info Align Int UNIQUE Int UNIQUE Int	
	     // for each ungapped block, parent_start child_start [length]
	     // starts are starts with respect to child (relevant if sequences 
	     //    are opposite orientations
	     // if no length then until next block (so no double gap) 
	     // if no Align assume ungapped starting at 1 in child
	   AlignDNAPep Int UNIQUE Int UNIQUE Int
           AlignPepDNA Int UNIQUE Int UNIQUE Int
             // These two tags are analogous to Align, but scale length
	     // for the case of a dna alignment to peptide or vice-versa.
           Mismatch Int UNIQUE Int	
	     // start end of mismatch region in child coords
	     // mismatches from this sequence or children are ignored
	     // if no end then only the specified base
	     // if no Ints then mismatches OK anywhere in this sequence
	   Contents  Homols_only
	             Features_only
		     No_DNA

in child object
	SMap  S_Parent <ptag2> <key>
	      S_Date UNIQUE DateType
	...
	Source_exons Int UNIQUE Int			// start end
	

note that if there are Source_exons, the final mapping should be the product of
that given in the parent and the child (not implemented yet 010323).
RD 010120 I'd like to phase out Source_exons
*/

/*******************************************************************/
/********** data structures ****************************************/

typedef struct {
  int s1, s2 ;	/* start/end of segments allowed to disagree, 
		   in source coordinates */
} SMapMismatch ;

struct SMapKeyInfoStruct {
  KEY key ;
  Array map ;			/* of SMapMap */
  Array mismatch ;		/* of SMapMisMatch */
  Array children ;		/* of SMapKeyInfo* */
} ;

struct SMapStruct {
  STORE_HANDLE handle ;		/* the SMap's own handle */
  Associator key2index ;	/* index into keyInfo */
  SMapKeyInfo *root ;
  int length ;
	/* following is a simple cache for key2index mapping */
  SMapKeyInfo *lastKeyInfo ;
} ;

/* For each key, the keyInfo record gives an array for the map and an
   array of mismatch information, which can be null. 

   Each SMapMap structure in map defines a set of intervals in the 
   coordinate system of key, each interval being mapped as an ungapped
   block to a corresponding interval in the SMap's coordinates.
   
   Each [s1,s2] pair in mismatch defines a set of intervals in the 
   coordinate system of key that can contain mismatches to the canonical 
   sequence.

   Both ->map and ->mismatch arrays are sorted in increasing order of s1, 
   and s2 >= s1 always.  We also assume here that the sequences are 
   colinear.
*/

/*********************************************************************/

static void sMapFinalize (void *ptr)
{ messfree (((SMap*) ptr)->handle) ;
}

static SMap* sMapSimpleCreate (STORE_HANDLE handle)
{
  SMap* smap = (SMap*) handleAlloc (sMapFinalize, handle, sizeof(SMap)) ;

  smap->handle = handleCreate () ;
  smap->key2index = assHandleCreate (smap->handle) ;

  return smap ;
}

void sMapDestroy (SMap* smap)
{ messfree (smap) ;		/* will call sMapFinalize */
}

/******************************************************************/

int sMapMax (SMap* smap) { return smap->length ; }

/******************************************************************/

SMapKeyInfo* sMapKeyInfo (SMap* smap, KEY key)
{
  SMapKeyInfo *info ;

  if (smap->lastKeyInfo && smap->lastKeyInfo->key == key)
    return smap->lastKeyInfo ;

  if (!assFind (smap->key2index, assVoid(key), &info))
    return 0 ;

  smap->lastKeyInfo = info ;
  return info ;
}

/******************************************************************/

/* binary searches to find segments used in sMapMap and sMapComposeMaps */

static SMapMap *lowestUpperSeg (Array map, float x)
{
  int i = 0, j = arrayMax(map)-1 ;
 
  while (i < j)
    if (x > arrp(map, (i+j)/2, SMapMap)->s2)
      i = (i+j)/2 + 1 ;
    else
      j = (i+j)/2 ;
  return arrp(map, i, SMapMap) ;
} 

static SMapMap *highestLowerSeg (Array map, float x)
{
  int i = 0, j = arrayMax(map)-1 ;
 
  while (i < j)
    if (x < arrp(map, (i+j)/2+1, SMapMap)->s1)
      j = (i+j)/2 ;
    else
      i = (i+j)/2 + 1 ;
  return arrp(map, j, SMapMap) ;
}

/******************************************************************/

/* fundamental coordinate conversion routine */

SMapStatus sMapMap (SMapKeyInfo* info, int x1, int x2, 
		    int *y1, int *y2, int *nx1, int *nx2)
{ 
  Array map = info->map ;
  int z1, z2 ;
  SMapMap *m1, *m2 ;
  SMapStatus status = 0 ;	/* default 0 is perfect map without gaps */
  BOOL isReverse = FALSE ;

		/* set default nx1, nx2 before possible reverse */
		/* this means they are set even if NO_OVERLAP, but that seems OK */
  if (nx1) *nx1 = x1 ; 
  if (nx2) *nx2 = x2 ;
  if (x1 > x2) { int t = x1 ; x1 = x2 ; x2 = t ; isReverse = TRUE ; }

  if (x1 > arrp(map, arrayMax(map)-1, SMapMap)->s2 || /* off end of last mapped segment */
      x2 < arrp(map, 0, SMapMap)->s1) /* off start of first mapped segment */
    return SMAP_STATUS_NO_OVERLAP_EXTERNAL ;

		/* find map segment for x1 by binary search */
		/* could cache in KeyInfo for rapid return on dense sorted features? */
  m1 = lowestUpperSeg (map, x1) ;

  if (x1 < m1->s1)		/* clip */
    { z1 = m1->r1 ;
      if (m1 == arrp(map, 0, SMapMap)) status |= SMAP_STATUS_X1_EXTERNAL_CLIP ;
      else status |= SMAP_STATUS_X1_INTERNAL_CLIP ;
      if (nx1 && !isReverse) *nx1 = m1->s1 ;
      else if (nx2 && isReverse) *nx2 = m1->s1 ;
    }
  else if (m1->r2 > m1->r1)
    z1 = m1->r1 + (x1 - m1->s1) ;
  else
    z1 = m1->r1 - (x1 - m1->s1) ;

		/* x2 likewise */
  m2 = highestLowerSeg (map, x2) ;

  if (m2 < m1)
    return SMAP_STATUS_NO_OVERLAP_INTERNAL ; /* completely contained in internal gap */
  if (m1 < m2)
    status |= SMAP_STATUS_INTERNAL_GAPS ;

  if (x2 > m2->s2)		/* clip */
    { z2 = m2->r2 ;
      if (m2 == arrp(map, arrayMax(map)-1, SMapMap))
	  status |= SMAP_STATUS_X2_EXTERNAL_CLIP ;
      else 
	  status |= SMAP_STATUS_X2_INTERNAL_CLIP ;
      if (nx2 && !isReverse) *nx2 = m2->s2 ;
      else if (nx1 && isReverse) *nx1 = m2->s2 ;
    }
  else if (m2->r2 > m2->r1)
    z2 = m2->r2 - (m2->s2 - x2) ;
  else
    z2 = m2->r2 + (m2->s2 - x2) ;

		/* done, now fill in results and return */
  if (isReverse) { if (y1) *y1 = z2 ; if (y2) *y2 = z1 ; }
  else { if (y1) *y1 = z1 ; if (y2) *y2 = z2 ; }
  return TRUE ;
}

/************************************************************/

/* standard coordinate conversion routine for a full SMap object */

BOOL sMapConvert (SMap* smap, KEY key, int x1, int x2, int *y1, int *y2)
{
  SMapKeyInfo *info = sMapKeyInfo (smap, key) ;
  SMapStatus status ;

  if (!info) 
    return FALSE ;

  status = sMapMap (info, x1, x2, y1, y2, 0, 0) ;
  return ((status & SMAP_STATUS_NO_OVERLAP) == 0) ;
}

/*************************************************************/

static void sMapKeyAdd (KEYSET kset, SMapKeyInfo *info)
{
  int i ;

  keySet(kset, keySetMax(kset)) = info->key ;
  for (i = 0 ; i < arrayMax (info->children) ; ++i)
    sMapKeyAdd (kset, arr(info->children, i, SMapKeyInfo*)) ;
}

KEYSET sMapKeys (SMap* smap, STORE_HANDLE handle)
{
  KEYSET kset = keySetHandleCreate (handle) ;

  sMapKeyAdd (kset, smap->root) ;

  return kset ;
}

/*************************************************************/

/* NB sMapDNA is called by dnaGet() if there is no directly
   attached DNA, and it calls dnaGet to get the dna for keys
   with directly attached DNA.  We have been careful to ensure 
   there is no loop here.
*/

static void sMapAddDNA (Array dna, SMapKeyInfo *info)
{
  KEY key = 0 ;
  OBJ obj = 0 ;
  Array a ;
  int i, j ;
  char *cp, *cq ;
  SMapMismatch *mis ;
  SMapMap *map ;

  if (bIndexFind (info->key, str2tag("DNA")) &&
      (obj = bsCreate (info->key)) && 
      bsGetKey (obj, str2tag("DNA"), &key) &&
      (a = dnaGet (key)))	/* simplest dnaGet() */
    { 

/* shift sequence allowed to mismatch into high byte */

      if (info->mismatch)
	for (i = 0 ; i < arrayMax (info->mismatch) ; ++i)
	  { mis = arrp(info->mismatch, i, SMapMismatch) ;
	    for (j = mis->s1 ; j < mis->s2 ; ++j)
	      arr(a,j,unsigned char) <<= 4 ;
	  }

/* strategy is to complain if low half-bytes don't match 
   then OR with new sequence, allowing high byte ambiguity codes */

      for (i = 0 ; i < arrayMax (info->map) ; ++i)
	{ map = arrp(info->map, i, SMapMap) ;
	  cp = arrp(dna, map->r1, unsigned char) ;
	  cq = arrp(a, map->s1, unsigned char) ;
	  if (map->r2 > map->r1)
	    for (j = map->s2 - map->s1 + 1 ; j-- ;)
	      { if ((*cp & 0xf) && (*cp & 0xf) != (*cq & 0xf))
		  messerror ("mismatch error at %d in %s", 
			     map->s2 - j, name(key)) ;
	        *cp++ |= *cq++ ;
	      }
	  else
	    for (j = map->s2 - map->s1 + 1 ; j-- ;)
	      { if ((*cp & 0xf) && (*cp & 0xf) != (*cq & 0xf))
		  messerror ("mismatch error at %d in %s", 
			     map->s2 - j, name(key)) ;
	        *cp-- |= *cq++ ;
	      }
	}

      arrayDestroy (a) ;
      bsDestroy (obj) ;
    }
  else 
    { if (obj) bsDestroy (obj) ;
				/* only recurse down if not added */
      for (i = 0 ; i < arrayMax (info->children) ; ++i)
	sMapAddDNA (dna, arr(info->children, i, SMapKeyInfo*)) ;
    }
}

Array sMapDNA (SMap* smap, STORE_HANDLE handle)
{
  Array dna ;		/* this is where we build the result */
  int i ;
  unsigned char *cp ;

  dna = arrayHandleCreate (smap->length, unsigned char, handle) ;

	/* recursively add DNA */

  sMapAddDNA (dna, smap->root) ;

	/* now fill in low byte from high byte if necessary*/

  cp = arrp(dna, 0, unsigned char) ;
  for (i = 0 ; i < arrayMax(dna) ; ++i)
    if (*cp & 0xf)
      *cp &= 0xf ;
    else
      *cp >>= 4 ;

  return dna ;
}

/*************************************************************/
/************ now the stuff to build smaps *******************/

static int SMapMapOrder (void *va, void *vb)
{
  SMapMap *a = (SMapMap*)va ;
  SMapMap *b = (SMapMap*)vb ;

  return (a->s1 - b->s1) ;
}

/******************************/

Array sMapLocalMap (OBJ obj,    /* in parent located at #SmapInfo */
		    int pStart, /* start coord in parent */
		    int pEnd,   /* end coord in parent */
		    STORE_HANDLE handle)   /* return array on this. */

{
  SMapMap *m ;
  Array u = arrayCreate (16, BSunit) ;
  BSunit *ui ;
  int i ;
  Array localMap ;
  int DNAfac, pepFac;
  BSMARK mark ;

  /* first get Align info into u */
  
  mark = bsMark (obj, 0) ;
  if (bsGetArray (obj, str2tag("Align"), u, 3))
    { DNAfac = 1 ; pepFac = 1 ; }
  else if (bsGetArray (obj, str2tag("AlignDNAPep"), u, 3))
    { DNAfac = 3 ; pepFac = 1 ; }
  else if (bsGetArray (obj, str2tag("AlignPepDNA"), u, 3))
    { DNAfac = 1 ; pepFac = 3 ; }
  else
    { arrayDestroy(u);
      bsGoto (obj, mark) ;
      bsMarkFree (mark) ;
      return 0;
    }
  bsGoto (obj, mark) ;
  bsMarkFree (mark) ;
  
  /* next convert u into localMap */
  localMap = arrayHandleCreate (arrayMax(u)/3, SMapMap, handle) ;
  for (i = 0 ; i < arrayMax(u)/3 ; ++i)
    { m = arrayp(localMap, i, SMapMap) ;
      ui = arrp(u, 3*i, BSunit) ;
      m->s1 = ui[1].i ;
      m->r1 = ui[0].i ;
      if (ui[2].i)
	 m->s2 = m->s1 + ui[2].i - 1 ;
    }
  arraySort (localMap, SMapMapOrder) ;

/** could/should check colinearity here? **/

	/* fill in blank ->s1 and all ->r2 values;  must come after sort */
  for (i = 0 ; i < arrayMax(localMap)-1 ; ++i)
    { m = arrp(localMap, i, SMapMap) ;
      if (pEnd > pStart)	/* forward alignment */
	{ if (!m->s2)
	    { if (((m+1)->s1 - m->s1) / pepFac < 
		  (((m+1)->r1 - m->r1) / DNAfac))
		m->s2 = (m+1)->s1 - 1 ;
	      else
		m->s2 = m->s1 + ((((m+1)->r1 - m->r1) / DNAfac) - 1) * pepFac ;
	    }
	  m->r2 = m->r1 + ((m->s2 - m->s1) * DNAfac) / pepFac ;
	}
      else			/* reversed alignment */
	{ if (!m->s2)
	    { if (((m+1)->s1 - m->s1) / pepFac < 
		  ((m->r1 - (m+1)->r1) / DNAfac))
		m->s2 = (m+1)->s1 - 1 ;
	      else
		m->s2 = m->s1 + (((m->r1 - (m+1)->r1) / DNAfac) - 1) * pepFac ;
	    }
	  m->r2 = m->r1 - ((m->s2 - m->s1) * DNAfac) / pepFac ;
	}
    }
  m = arrp(localMap, arrayMax(localMap)-1, SMapMap) ;
  m->r2 = pEnd ;
  if (!m->s2)
    { if (pEnd > pStart)
	m->s2 = m->s1 + ((m->r2 - m->r1) / DNAfac) * pepFac ;
      else
	m->s2 = m->s1 + ((m->r1 - m->r2) / DNAfac) * pepFac ;
    }
/** could check total lengths match here if given **/

  arrayDestroy (u) ;
  return localMap ;
}

/******************************/

static Array sMapComposeMaps (Array localMap, SMapKeyInfo *parentInfo,
			      STORE_HANDLE handle)
{ 
  Array productMap ;
  SMapMap *m1, *m2, *m3 ;
  SMapStatus status ;
  int i, y1, y2, z1, z2 ;

/* Warning: a picture helps A LOT if you want to follow this function */

/* Notation:
   m1 in localMap maps x-space to y-space
   m2 in parentMap maps y-space to z-space
   m3 in the productMap maps x-space to z-space
*/

  productMap = arrayHandleCreate (arrayMax(localMap) + 
				  arrayMax(parentInfo->map) - 1, 
				  SMapMap, handle) ;

/* First loop over localMap, creating 0, 1 or more segments in
   productMap for each localMap segment.  Treat separately the two
   cases where localMap is reversed or not reversed.
*/

  if (arrp(localMap,0,SMapMap)->r1 < arrp(localMap,arrayMax(localMap)-1,SMapMap)->r2)
    				/* localMap is not reversed - easier to think about */
    for (i = 0 ; i < arrayMax(localMap) ; ++i)
      { m1 = arrp(localMap, i, SMapMap) ;
        status = sMapMap (parentInfo, m1->r1, m1->r2, &z1, &z2, &y1, &y2) ;
	if (status & SMAP_STATUS_NO_OVERLAP) continue ;

	m3 = arrayp(productMap, arrayMax(productMap), SMapMap) ; /* make at least one segment */

	if (!(status & SMAP_STATUS_INTERNAL_GAPS)) /* easy case */
	  { m3->s1 = m1->s1 + (y1 - m1->r1) ; m3->r1 = z1 ;
	    m3->s2 = m1->s2 - (m1->r2 - y2) ; m3->r2 = z2 ;
	  }
	else			/* complex case when INTERNAL_GAPS */
	  { m3->s1 = m1->s1 + (y1 - m1->r1) ; m3->r1 = z1 ; /* start first segment as above */

	    m2 = lowestUpperSeg (parentInfo->map, m1->r1) ;
	    while (m2->s2 < y2) /* all the rest of m2 is in the m1 alignment */
	      { m3->s2 = m1->s1 + (m2->s2 - m1->r1) ; m3->r2 = m2->r2 ; /* finish this segment */
		++m2 ;
		m3 = arrayp(productMap, arrayMax(productMap), SMapMap) ; /* and start another */
		m3->s1 = m1->s1 + (m2->s1 - m1->r1) ; m3->r1 = m2->r1 ;
	      }
	    m3->s2 = m1->s2 - (m1->r2 - y2) ; m3->r2 = z2 ;	/* finish final segment */
	  }
      }
  else				/* localMap is reversed */
    for (i = 0 ; i < arrayMax(localMap) ; ++i)
      { m1 = arrp(localMap, i, SMapMap) ;
        status = sMapMap (parentInfo, m1->r1, m1->r2, &z1, &z2, &y1, &y2) ;
	if (status & SMAP_STATUS_NO_OVERLAP) continue ;

	m3 = arrayp(productMap, arrayMax(productMap), SMapMap) ; /* make at least one segment */

	if (!(status & SMAP_STATUS_INTERNAL_GAPS)) /* easy case */
	  { m3->s1 = m1->s1 + (m1->r1 - y1) ; m3->r1 = z1 ;
	    m3->s2 = m1->s2 - (y2 - m1->r2) ; m3->r2 = z2 ;
	  }
	else			/* complex case when INTERNAL_GAPS */
	  { m3->s1 = m1->s1 + (m1->r1 - y1) ; m3->r1 = z1 ; /* start first segment as above */
	    m2 = highestLowerSeg (parentInfo->map, m1->r1) ;
	    while (m2->s1 > y2) /* all the rest of m2 is in the m1 alignment */
	      { m3->s2 = m1->s1 + (m1->r1 - m2->s2) ; m3->r2 = m2->r2 ; /* finish this segment */
		--m2 ;
		m3 = arrayp(productMap, arrayMax(productMap), SMapMap) ; /* and start another */
		m3->s1 = m1->s1 + (m1->r1 - m2->s1) ; m3->r1 = m2->r1 ;
	      }
	    m3->s2 = m1->s2 - (y2 - m1->r2) ; m3->r2 = z2 ;	/* finish final segment */
	  }
      }

/* finally must go through and check that we don't have abutting segments - if so join them */
/* reuse m1, m2, m3 all in productMap */

  m2 = m1 = arrp(productMap, 0, SMapMap) ;
  for (i = 0 ; i < arrayMax(productMap)-1 ; ++m2, ++i)
    { m3 = m2 + 1 ;
      if (m3->s1 > m2->s2 + 1 || m3->r1 > m2->r2 + 1 || m3->r1  < m2->r2 - 1)
        { if (m1 != m2) { m1->s2 = m2->s2 ; m1->r2 = m2->r2 ; }
	  ++m1 ;
	  if (m1 != m3) { m1->s1 = m3->s1 ; m1->r1 = m3->r1 ; }
	}
    }
  if (m1 != m2) { m1->s2 = m2->s2 ; m1->r2 = m2->r2 ; }
  arrayMax(productMap) = (m1 - arrp(productMap, 0, SMapMap)) + 1 ;

  return productMap ;
}

/******************************/

/* a utility I thought I would need, but ended up not doing */
/* but it might be useful one day, so I leave it */

void sMapInvertMap (Array map)
{
  int i, t ;
  SMapMap *m, *n ;

  if (arrp(map, 0, SMapMap)->r1 < arrp(map, 0, SMapMap)->r2)
    for (i = 0 ; i < arrayMax(map) ; ++i)
      { m = arrp(map, i, SMapMap) ;
        t = m->s1 ; m->s1 = m->r1 ; m->r1 = t ;
        t = m->s2 ; m->s2 = m->r2 ; m->r2 = t ;
      }
  else
    for (i = 0 ; i < arrayMax(map) ; ++i)
      { m = arrp(map, i, SMapMap) ;
        n = arrp(map, arrayMax(map)-1-i, SMapMap) ;
        t = m->s1 ; m->s1 = n->r2 ; n->r2 = t ;
        t = m->s2 ; m->s2 = n->r1 ; n->r1 = t ;
      }
}

/*********************************************************/

static SMapKeyInfo *sMapNewInfo (SMap* smap, KEY key)
{
  SMapKeyInfo *info = (SMapKeyInfo*) halloc (sizeof(SMapKeyInfo),
					     smap->handle) ;
  info->key = key ;
  assInsert (smap->key2index, assVoid(key), info) ;
  info->children = arrayHandleCreate (8, SMapKeyInfo*, smap->handle) ;

  return info ;
}

/*****************/

static void exonsSort (Array a) ; /* forward declaration (see EOF) */

static BOOL sMapAddLocal (SMap *smap, 
			  KEY parent,
			  KEY key,    /* key to add */
			  int pStart, /* start coord in parent */
			  int pEnd,   /* end coord in parent */
			  OBJ parentObj) /* located at #SmapInfo */
{
  SMapKeyInfo *info, *parentInfo ;
  Array localMap, u ;
  SMapMismatch *mis ;
  SMapMap *m ;
  int i ;

  if (assFind (smap->key2index, assVoid(key), 0)) /* we have it already */
    return FALSE ;

  info = sMapNewInfo (smap, key) ;
  parentInfo = sMapKeyInfo(smap, parent) ;
  array(parentInfo->children, arrayMax(parentInfo->children),
	SMapKeyInfo*) = info ;

	/* make local map then compose with parent map to give map */

  if (!(localMap = sMapLocalMap(parentObj, pStart, pEnd, 0)))
    {
      localMap = arrayCreate (1, SMapMap) ;
      m = arrayp (localMap, 0, SMapMap) ;
      m->s1 = 1 ;
      m->s2 = (pEnd > pStart) ? (pEnd - pStart + 1) : (pStart - pEnd + 1) ;
      m->r1 = pStart ;
      m->r2 = pEnd ;
    }
  
  info->map = sMapComposeMaps (localMap, parentInfo, smap->handle) ;
  arrayDestroy (localMap) ;

#ifdef ACEDB4	/* must deal with Source_exons also as a localMap */
  { OBJ obj = 0 ;
    int x ;

    u = arrayCreate (16, BSunit) ;
    if (bIndexFind (key, str2tag("Source_exons")) &&
	(obj = bsCreate(key)) &&
	bsGetArray (obj, str2tag("Source_exons"), u, 2))
      { exonsSort (u) ;
        localMap = arrayCreate (arrayMax(u)/2, SMapMap) ;
	x = 1 ;
	for (i = 0 ; i < arrayMax(u) ; i += 2)
	  { m = arrayp (localMap, i/2, SMapMap) ;
	    m->r1 = arr(u,i,BSunit).i ;
	    m->r2 = arr(u,i+1,BSunit).i ;
	    m->s1 = x ;
	    m->s2 = x + (m->r2 - m->r1) ;
	    x += m->r2 - m->r1 + 1 ;
	  }
	{ Array temp = sMapComposeMaps (localMap, info, smap->handle) ;
	  arrayDestroy (info->map) ;
	  info->map = temp ;
	}
	arrayDestroy (localMap) ;
      }
    if (obj) bsDestroy (obj) ;
    arrayDestroy (u) ;
  }
#endif /* ACEDB4 */

	/* get and store info->mismatch */

  info->mismatch = arrayHandleCreate (arrayMax(u)/2, SMapMismatch,
				      smap->handle) ;
  u = arrayCreate (16, BSunit) ;
  if (bsGetArray (parentObj, str2tag("Mismatch"), u, 2))
    for (i = 0 ; i < arrayMax(u) ; i += 2)
      { mis = arrayp(info->mismatch, arrayMax(info->mismatch), SMapMismatch) ;
        mis->s1 = arr(u,i,BSunit).i ;
	mis->s2 = arr(u,i+1,BSunit).i ? arr(u,i+1,BSunit).i : mis->s1 ;
      }
  else if (bsFindTag (parentObj, str2tag("Mismatch"))) /* whole sequence */
    { mis = arrayp(info->mismatch, 0, SMapMismatch) ;
      mis->s1 = arrp(info->map, 0, SMapMap)->s1 ;
      mis->s2 = arrp(info->map, arrayMax(info->map), SMapMap)->s2 ;
    }
  arrayDestroy (u) ;
  if (!arrayMax (info->mismatch))
    arrayDestroy (info->mismatch) ;

  return TRUE ;
}

static void sMapAddChildren (SMap *smap, KEY key, char *aqlCondition)
		/* aqlCondition does nothing yet */
{
  int pos1, pos2 ;
  BSMARK mark1 = 0, mark2 = 0 ;
  OBJ obj ;
  KEY child ;

	/* check if any children first before opening */
  if (!bIndexFind (key, str2tag("S_Child"))
#ifdef ACEDB4
      && !bIndexFind (key, str2tag("Subsequence"))
#endif
      ) return ;

  if (!(obj = bsCreate (key)))
    return ;

  if (bsGetKeyTags (obj, str2tag("S_Child"), 0)) do
    { mark1 = bsMark (obj, mark1) ;
      if (bsGetKey (obj, _bsRight, &child)) do
	{ mark2 = bsMark (obj, mark2) ;
	  if (bsGetData (obj, _bsRight, _Int, &pos1) &&
	      bsGetData (obj, _bsRight, _Int, &pos2) &&
	      sMapConvert (smap, key, pos1, pos2, 0, 0))
/* sMapConvert() call checks this key overlaps area of smap */
	    { bsPushObj (obj) ; 
	      if (sMapAddLocal (smap, key, child, pos1, pos2, obj))
		sMapAddChildren (smap, child, aqlCondition) ;
	    }
	  bsGoto (obj, mark2) ;
	} while (bsGetKey (obj, _bsDown, &child)) ;
      bsGoto (obj, mark1) ;
    } while (bsGetKeyTags (obj, _bsDown, 0)) ;

#ifdef ACEDB4	/* support Subsequence not under S_Child temporarily */
  bsGoto (obj, 0) ;
  if (bsGetKey (obj, str2tag("Subsequence"), &child)) do
    { mark1 = bsMark (obj, mark1) ;
      if (bsGetData (obj, _bsRight, _Int, &pos1) &&
	  bsGetData (obj, _bsRight, _Int, &pos2) &&
	  sMapConvert (smap, key, pos1, pos2, 0, 0))
	{ bsPushObj (obj) ; 
	  if (sMapAddLocal (smap, key, child, pos1, pos2, obj))
	    sMapAddChildren (smap, child, aqlCondition) ;
	}
      bsGoto (obj, mark1) ;
    } while (bsGetKey (obj, _bsDown, &child)) ;
#endif /* ACEDB4 */
  
  bsMarkFree (mark1) ; bsMarkFree (mark2) ;
  bsDestroy (obj) ;
}

SMap* sMapCreate (STORE_HANDLE handle, KEY key, int x1, int x2,
		  char *aqlCondition)
	/* key is top level key */
	/* aqlCondition does nothing yet */
{
  SMap *smap = sMapSimpleCreate (handle) ;
  SMapMap *m ;

/* insert self into smap as root */

  smap->root = sMapNewInfo (smap, key) ;
  smap->root->map = arrayCreate (1, SMapMap) ;
  m = arrayp(smap->root->map, 0, SMapMap) ;
  if (x2 > x1)
    { m->s1 = x1 ; m->s2 = x2 ;
      m->r1 = 1 ; m->r2 = x2 - x1 + 1 ;
    }
  else
    { m->s1 = x2 ; m->s1 = x1 ;
      m->r1 = x1 - x2 + 1 ; m->r2 = 1 ;
    }

  smap->length = (x2 > x1) ? (x2 - x1 + 1) : (x1 - x2 + 1) ;

/* now recursively add all children */

  sMapAddChildren (smap, key, aqlCondition) ;

  return smap ;
}

/************************************************************************/
/******** routines to be used independently/prior to SMAP creation ******/

BOOL sMapEnclosedBy (KEY key, int x1, int x2, 
		     KEY *kp, int *y1, int *y2)
  /* finds next enclosing sequence, with coords in that sequence */
{
  OBJ obj ;
  KEY parent ;
  int z1, z2 ;
  BOOL result ;
  SMapKeyInfo info ;	/* local structure so can send right argument to sMapMap() */

  if (x1 <= 0 || x2 < 0)	/* x2 == 0 means set to length */
    return FALSE ;

  if (!bIndexFind (key, str2tag("S_Parent"))
#ifdef ACEDB4
      && !bIndexFind (key, str2tag("Source"))
#endif
      ) return FALSE ;

  if (!(obj = bsCreate (key))) 
    return FALSE ;

      /* find self in parent */
      /* slightly convoluted to allow use of Source #ifdef ACEDB4 */

  parent = 0;
  if (bsGetKeyTags (obj, str2tag("S_Parent"), 0))
    bsGetKey (obj, _bsRight, &parent) ;
  bsDestroy (obj) ;
  if (!(parent && (obj = bsCreate (parent)) &&
	bsFindKey2 (obj, str2tag("S_Child"), key, 0) &&
	bsGetData (obj, _bsRight, _Int, &z1) &&
	bsGetData (obj, _bsRight, _Int, &z2)))
    {
      if (obj) bsDestroy (obj) ;

#ifndef ACEDB4
      return FALSE;
#else	/* accept Source not under S_Parent temporarily */
      obj = bsCreate (key) ;
      parent = 0;
      bsGetKey (obj, str2tag("Source"), &parent);
      bsDestroy (obj) ;
      if (!(parent && (obj = bsCreate (parent)) &&
	    bsFindKey (obj, str2tag("Subsequence"), key) &&
	    bsGetData (obj, _bsRight, _Int, &z1) &&
	    bsGetData (obj, _bsRight, _Int, &z2)))
	{
	  if (obj) bsDestroy (obj) ;
	  return FALSE; 
	}
#endif /* ACEDB4 */
    }

  bsPushObj (obj) ;		/* go into #SMap_info */
  if (!(info.map = sMapLocalMap (obj, z1, z2, 0))) 
    {
      SMapMap *m ;
      info.map = arrayCreate (1, SMapMap) ;
      m = arrayp (info.map, 0, SMapMap) ;
      m->r1 = z1 ;
      m->r2 = z2 ;
      m->s1 = 1 ;
      m->s2 = (z2 > z1) ? (z2 - z1 + 1) : (z1 - z2 + 1) ;
    }
  bsDestroy (obj) ;
  
				/* convert x2 == 0 to length */
  if (x2 == 0) 
    x2 = (z2 > z1) ? (z2 - z1 + 1) : (z1 - z2 + 1) ;

  if (sMapMap (&info, x1, x2, &z1, &z2, 0, 0)) /* reuse z1, z2 */
    { if (kp) *kp = parent ;
      if (y1) *y1 = z1 ;
      if (y2) *y2 = z2 ;
      result = TRUE ;
    }
  else
    result = FALSE ;

  arrayDestroy (info.map) ;
  return result ;
}

/*****************/

BOOL sMapTreeRoot (KEY key, int x1, int x2,
		   KEY *kp, int *y1, int *y2)
{
  static KEY k ;
  static int k1, k2 ;
  static Associator hashSet = 0 ;

				/* use hashSet to break loops */
  hashSet = assReCreate (hashSet) ;

				/* need non-zero kp, y1, y2 */
  if (!kp) kp = &k ;
  if (!y1) y1 = &k1 ;
  if (!y2) y2 = &k2 ;

  assInsert (hashSet, assVoid(*kp), assVoid(1)) ;
  if (sMapEnclosedBy (key, x1, x2, kp, y1, y2))
    { while (assInsert (hashSet, assVoid(*kp), assVoid(1)) &&
	     sMapEnclosedBy (*kp, *y1, *y2, kp, y1, y2)) ;
      return TRUE ;
    }
  else
    return FALSE ;
}

/*****************/

/* try various strategies to find the sequence length of an object
   could do all and check consistency?
*/

int sMapLength (KEY key)
{
  static Array units = 0 ;
  int i, max = 0 ;
  int z1, z2 ;
  OBJ obj ;

  units = arrayReCreate (units, 48, int) ;

  if (!bIndexFind (key, str2tag("DNA")) &&
      !bIndexFind (key, str2tag("S_Parent")) &&
      !bIndexFind (key, str2tag("S_Child"))
#ifdef ACEDB4
      && !bIndexFind (key, str2tag("Source"))
      && !bIndexFind (key, str2tag("Subsequence"))
#endif
      ) max = 0 ;

  else if (!(obj = bsCreate (key)))
    max = 0 ;

	/* check first if it has its own DNA */

  else if (bsGetKey (obj, str2tag("DNA"), 0) &&
      bsGetData (obj, _bsRight, _Int, &max))
    max = max ;

	/* next if S_parent */

  else if (sMapEnclosedBy (key, 1, 0, 0, &z1, &z2))
    max = (z2 > z1) ? (z2 - z1 + 1) : (z1 - z2 + 1) ;

	/* next if S_children */

  else if (bsGetArray (obj, str2tag("S_Child"), units, 4))
    for (i = 2 ; i < arrayMax (units) ; i += 4)
      { if (arr(units,i,BSunit).i > max)
	  max = arr(units,i,BSunit).i ;
	if (arr(units,i+1,BSunit).i > max)
	  max = arr(units,i+1,BSunit).i ;
      }

	/* or Subsequence in ACEDB 4 */
#ifdef ACEDB4
  else if (bsGetArray (obj, str2tag("Subsequence"), units, 3))
    for (i = 1 ; i < arrayMax (units) ; i += 3)
      { if (arr(units,i,BSunit).i > max)
	  max = arr(units,i,BSunit).i ;
	if (arr(units,i+1,BSunit).i > max)
	  max = arr(units,i+1,BSunit).i ;
      }
#endif /* ACEDB4 */

  bsDestroy (obj) ;
  return max ;
}

/***********************************************************/
/******* a utility needed for Source_exons sorting *********/

typedef struct { int x, y ; } ExonStruct ;

static int dnaExonsOrder (void *a, void *b)
{
  ExonStruct *ea = (ExonStruct *)a ;
  ExonStruct *eb = (ExonStruct *)b ;
  return ea->x - eb->x ;
}

static void exonsSort (Array a)
{		/* can't just arraySort, because a is in BSunits, not
		   pairs of BSunits */
  int i, n ;
  Array b = 0 ;

  n = arrayMax(a) / 2 ;
  b = arrayCreate (n, ExonStruct) ;
  for (i = 0 ; i < n ; ++i)
    { arrayp(b,i,ExonStruct)->x = arr(a,2*i,BSunit).i ;
      arrp(b,i,ExonStruct)->y = arr(a,2*i+1,BSunit).i ;
    }
  arraySort (b, dnaExonsOrder) ;
  for (i = 0 ; i < n ; ++i)
    { arr(a,2*i,BSunit).i = arrp(b,i,ExonStruct)->x ;
      arr(a,2*i+1,BSunit).i = arrp(b,i,ExonStruct)->y ;
    }
  arrayDestroy (b) ;
}

/*********************** end of file ************************/
