/*  File: fmapcurate.c
 *  Author: Richard Durbin (rd@sanger.ac.uk)
 *  Copyright (C) Richard Durbin, Genome Research Limited 2001
 * -------------------------------------------------------------------
 * Acedb is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * or see the on-line version at http://www.gnu.org/copyleft/gpl.txt
 * -------------------------------------------------------------------
 * This file is part of the ACEDB genome database package, written by
 * 	Richard Durbin (Sanger Centre, UK) rd@sanger.ac.uk, and
 *	Jean Thierry-Mieg (CRBM du CNRS, France) mieg@kaa.cnrs-mop.fr
 *
 * Description: Curation tools to accept new CDS
 * HISTORY:
 * Last edited: Jun 27 15:32 2001 (rd)
 * Created: Jun 27 14:58 2001 (rd)
 * $Id: fmapcurate.c,v 1.1 2001/07/23 14:21:51 rd Exp $
 *-------------------------------------------------------------------
 */

  /* Strategy: make window with following:

if cds overlaps exactly an existing cds
  edit remark for CDS xyz.n

if cds overlaps nothing:
  create new gene xyz.n

if cds overlaps one gene and disagrees:
  [] create new splice form xyz.na for gene xyz.n
  [] replace existing CDS xyz.na <for each existing cds>
  [] create new gene xyz.m split from xyz.n
forced choice of one of these

if cds overlaps multiple genes:
 merge and create new cds with gene name
      []  xyz.n <for each gene overlapped>
 replace  
   new name  bury
      []      []   xyz.na <for each cds of each gene overlapped>
 
only one of merge, replace/new name.  Can bury many if replacing.
    
Remark buffer preset with:
  [<date> <name>] 
If choices required, then must complete these first, which auto-sets initial text in remark buffer and puts cursor at end.

at bottom two buttons: [CURATE] [CANCEL]
   */
  
#include "acedb.h"
#include "aceio.h"
#include "lex.h"
#include "main.h"		/* for checkWriteAccess() */
#include "query.h"
#include <w7/fmap_.h>

static int CURATE_STATE_HOOK ;

typedef struct {
  FeatureMap fmap ;
  Graph graph ;			/* for the curation window */
  STORE_HANDLE handle ;
  KEY key ;			/* key of CDS object to curate */
  Array cdsOverlaps ;
  Array geneOverlaps ;
  char *newGeneName ;
  char *newCDSname ;
  KEY exactMatch ;		/* set by curateExactMatch() */
  int option ;			/* which action to take */
  char *remarkText ;
} CurateState ;

static BOOL curateOverlaps (CurateState *state) ;
static void newGeneName (CurateState *state) ;
static BOOL curateExactMatch (CurateState *state) ;
static void curateSubmit (void) ;

/******************************************************/

void fMapCurateWindow (int box)
{
  CurateState *state ;
  FeatureMap look = currentFeatureMap ("fMapCurateWindow") ;
  int y ;
  static char *user ;

  if (!checkWriteAccess ()) 
    return ;
  if (!(user = messPrompt ("Give user name for remarks (first time only)", 
			     getLogin(TRUE), "wz", 0)))
    return ;

  graphCreate (TEXT_SCROLL, "CDS overlap curation", 0, 0, 0.4, 0.4) ;
  state = (CurateState*) halloc (sizeof(CurateState), graphHandle()) ;
  graphAssociate (&CURATE_STATE_HOOK, state) ;

  state->key = BOXSEG(box)->parent ;
  state->fmap = look ;		/* must get before graphCreate() */
  state->graph = graphActive() ;
  state->handle = graphHandle() ;

	/* make ->cdsOverlaps and ->geneOverlaps */
  if (!curateOverlaps (state))
    { messout ("Sorry, I can't find CDS %s on this display") ;
      graphDestroy () ;		/* destroys state automatically */
      return ;
    }

  /* now set up options */

  state->option = 0 ;		/* do nothing - just add remark */
  y = 1 ;
  if (curateExactMatch (state))
    { graphText (hprintf (state->handle,
			  "Exact match to curated CDS %s",
			  name (state->exactMatch)), 
		 2, y ) ; 
      y += 1.5 ;
      if (state->key == state->exactMatch) /* this is the curated CDS */
	{ graphToggleEditor ("Kill this gene", &state->option, 2, y) ;
	  y += 1.5 ;
	}
    }

  else if (!keySetMax (state->cdsOverlaps))
    { curateNewGeneName (state) ;
      graphText (hprintf (state->handle,
			  "No overlapping curated CDS.  "
			  "Create new gene %s",
			  name (state->newCDSname)),
		 2, y) ;
      y += 1.5 ;
    }

  else if (keySetMax (state->geneOverlaps) == 1)
    { int tag ;
      KEY gene = keySet (state->geneOverlaps, 0) ;

      graphText (hprintf (state->handle, 
			  "Overlap to gene %s", 
			  name(gene)), 
		 2, y) ; 
      y += 1.5 ;
      tag = graphRadioCreate ("Options:", &state->option, 2, y++) ;
      curateNewCDSname (state, gene) ;
      graphAddRadioEditor ("Add new CDS %s", tag, 2, 3, y++) ;

      curateNewGeneName (state) ;
      graphAddRadioEditor ("Split off new gene %s", tag, 3, 3, y++) ;
      graphAddRadioEditor ("Replace CDS", tag, 4, 3, y++) ;
      state->option = 1 ;	/* default is to add new CDS */
      graphSetRadioEditor (tag, state->option) ;
    }

  else				/* potentially merge genes */
    { 
    }

  state->remarkText = (char*) halloc (1024, state->handle) ;
  sprintf (state->remarkText, "[%s %s] ", 
	   timeShow (timeParse ("today")), user) ;
  graphTextScrollEditor ("Remark:", state->remarkText, 1024, 40,
			 1, y, 0) ;
  y += 1.5 ;

  graphButton ("Submit", curateSubmit, 3, y) ;
  graphButton ("Cancel", graphDestroy, 10, y) ;
  
  return ;
}

/****************************************************/

static void curateSubmit (void) 
{
  CurateState *state ;
  OBJ obj ;
  KEY key ;

  if (!graphAssFind (&CURATE_STATE_HOOK, &state))
    messcrash ("Can't find state in curateSubmit()") ;

	/* make sure all editor values are read back */
  graphCheckEditors (graphActive(), FALSE) ;


  if (!graphActivate (state->fmap->graph))
    {
      messout ("The parent FMAP graph is no longer present - aborting") ;
      goto end ;
    }

  if (!(obj = bsUpdate (state->key)))
    { messout ("Could not update the primary object - aborting") ;
      graphDestroy () ;
      return ;
    }
  lexaddkey (state->remarkText, &key, pickWord2Class ("Text")) ;
  bsAddKey (obj, str2tag("Remark"), key) ;
  bsSave (obj) ;

  switch (state->option)
    { 
    case 0: /* do nothing */
      break ;
    case 1: /* kill a curated CDS */
      if ("this is the only CDS for its gene" &&
          !messquery ("This is the only CDS for gene %s.  Are you sure you want to kill this, and hence (possibly) the gene?"))
	graphDestroy () ;
      add ("a remark to the parent about removing this CDS") ;
      makeHistory (state->key, TRUE) ; /* TRUE renames object, deleting Gene and Transcript links and resetting the method */
      break ;
    case 2:			/* add new CDS */
      key = fixCDS (state->key, state->newCDSname) ;
      add ("key to existing parent gene, with remark") ;
      break ;
    case 3:			/* split off new gene */
      key = fixCDS (state->key, state->newGeneName) ;
      create ("new gene and add key to it, with remark") ;
      break ;
    default: 
    }

 end:
  graphActivate (state->graph) ;
  graphDestroy () ;		/* frees all memory */
}

/****************************************************/

static BOOL curateOverlaps (CurateState *state)
{
  int min, max ;
  int i, j ;
  BOOL foundSomething = FALSE ;
  Seg *iseg, *jseg ;
  KEY curatedKey ;
  KEYSET cdsSet ;

  if (!lexClassKey ("Method:curated", &curatedKey))
    return FALSE ;
  if (!fMapFindSegBounds (state->fmap, EXON, &min, &max)) 
    return FALSE ;

  state->cdsOverlaps = cdsSet = keySetHandleCreate (state->handle) ;

/* loop over EXON segs to find exons for state->key, and then for
   each one find overlapping exons with method "curated" */

  iseg = arrp(state->fmap->segs, min, SEG) ;
  for (i = min ; i < max ; ++i, ++iseg)
    if (iseg->parent == state->key)
      { foundSomething = TRUE ;
        jseg = arrp(state->fmap->segs, min, SEG) ;
        for (j = min ; j < max ; ++j, ++jseg)
	  if (jseg->x1 < iseg->x2 && jseg->x2 > iseg->x1 &&
	      jseg->parent != state->key &&
	      lexAliasOf(arrp(state->fmap->seqInfo, jseg->data.i, SEQINFO)->method) == curatedKey) /* overlap */
	    keySet(cdsSet,keySetMax(cdsSet)) = jseg->parent ;
      }

/* now make the list unique, find the genes that these correspond to, and sort alphabetically */

  keySetSort (cdsSet) ;
  keySetCompress (cdsSet) ;
  state->geneOverlaps = query (cdsSet, "FOLLOW Gene") ;
  arraySort (cdsSet, keySetAlphaOrder) ;
  arraySort (state->geneOverlaps, keySetAlphaOrder) ;

  if (keySetMax (cdsSet) && !keySetMax (state->geneOverlaps))
    return FALSE ;

  return foundSomething ;
}

/****************************************************/

static void newGeneName (CurateState *state)
{
  
}

/********* end of file ***********/
