/*  File: smap.h
 *  Author: Richard Durbin (rd@sanger.ac.uk)
 *  Copyright (c) J Thierry-Mieg and R Durbin, 1999
 * -------------------------------------------------------------------
 * Acedb is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * or see the on-line version at http://www.gnu.org/copyleft/gpl.txt
 * -------------------------------------------------------------------
 * This file is part of the ACEDB genome database package, written by
 * 	Richard Durbin (Sanger Centre, UK) rd@sanger.ac.uk, and
 *	Jean Thierry-Mieg (CRBM du CNRS, France) mieg@kaa.crbm.cnrs-mop.fr
 *
 * Description: 
 * HISTORY:
 * Last edited: Jan 29 20:20 2001 (rd)
 * Created: Wed Jul 28 22:11:39 1999 (rd)
 * CVS info:   $Id: smap.h,v 1.11 2001/08/29 09:29:32 rd Exp $
 *-------------------------------------------------------------------
 */

#ifndef DEF_SMAP_H
#define DEF_SMAP_H

/* #define USE_SMAP  define this to use new smap code */

#include "wh/regular.h"
#include "wh/bs.h"

typedef struct {
  int s1, s2 ;			/* coords in input space */
  int r1, r2 ;			/* coords in output space */
} SMapMap ;

typedef struct SMapStruct SMap ;

SMap* sMapCreate (STORE_HANDLE handle, KEY key, int x1, int x2,
		  char *aqlCondition) ;
  /* Recursively builds map down from key between start and stop inclusive
     in key's coordinate system. 
     The idea is that the aqlCondition is applied to test whether to open 
     each object, but this is not implemented yet.
     NB x2 can not be 0 here: must explicitly get length first
  */

Array sMapLocalMap (OBJ obj,    /* in parent located at #SmapInfo */
		    int pStart, /* start coord in parent */
		    int pEnd,   /* end coord in parent */
		    STORE_HANDLE h);   /* return array on this. */

void sMapDestroy (SMap* smap) ;

int sMapMax (SMap* smap) ;	/* sMapLength() used below for a key */

/* basic coordinate conversion */

BOOL sMapConvert (SMap* smap, KEY key, int x1, int x2, int *y1, int *y2) ;
  /* Converts interval x1..x2 in key's coordinate system to y1..y2 in
     smap's.  Returns TRUE if succeeds, FALSE if e.g. key not known in smap,
     or both x1, x2 out of range (on the same side), or both fall in the same
     gap in the map.  If part of the interval is out of range then it clips. 
  */

/* The next section provides more complex support for coordinate conversion. */
/* These are called by sMapConvert(), so are also more efficient. */

typedef struct SMapKeyInfoStruct SMapKeyInfo ;

SMapKeyInfo* sMapKeyInfo (SMap* smap, KEY key) ; /* gives handle for use with SMapMap */

typedef unsigned int SMapStatus ; /* status == 0 for clean map, else some bits below set */

#define SMAP_STATUS_NO_OVERLAP_EXTERNAL 0x0001 /* failure: [x1,x2] outside map range */
#define SMAP_STATUS_NO_OVERLAP_INTERNAL 0x0002 /* failure: [x1,x2] within a gap in map */
#define SMAP_STATUS_NO_OVERLAP		0x0003 /* either of the above: y1, y2 unchanged */
#define SMAP_STATUS_INTERNAL_GAPS	0x0004 /* gap(s) within alignment */
#define SMAP_STATUS_X1_EXTERNAL_CLIP	0x0010 /* x1 off end of map */
#define SMAP_STATUS_X1_INTERNAL_CLIP	0x0020 /* x1 in internal gap in map */
#define SMAP_STATUS_X1_CLIP		0x0030 /* either of the above: implies *nx1 != x1 */
#define SMAP_STATUS_X2_EXTERNAL_CLIP	0x0040 /* etc. */
#define SMAP_STATUS_X2_INTERNAL_CLIP	0x0080
#define SMAP_STATUS_X2_CLIP		0x00c0
#define SMAP_STATUS_CLIP		0x00f0

SMapStatus sMapMap (SMapKeyInfo* info, int x1, int x2, 
		    int *y1, int *y2, int *nx1, int *nx2) ;
  /* x1, x2, y1, y2 as sMapConvert() */
  /* return value is status as above */
  /* nx1, nx2 are the clipped values that actually map to y1, y2 - for code simplicity 
     these are always set, even if NO_OVERLAP status is returned (in which case they 
     are x1, x2) */

KEYSET sMapKeys (SMap* smap, STORE_HANDLE handle) ;
  /* this gives all keys that map onto this smap */

Array sMapDNA (SMap* smap, STORE_HANDLE handle) ;
  /* gets the DNA in acedb units */

/* The remaining functions don't need SMap structure - they can be used
   independently, e.g. sMapTreeRoot() to find the ultimate ancestor prior to 
   calling sMapCreate().
*/

BOOL sMapEnclosedBy (KEY key, int x1, int x2, 
		     KEY *kp, int *y1, int *y2) ;
  /* finds next enclosing sequence, with coords in that sequence */
  /* returns FALSE if no enclosing object, or can't map coords into it */
  /* x2 == 0 means end of sequence */

BOOL sMapTreeRoot (KEY key, int x1, int x2,
		   KEY *kp, int *y1, int *y2) ;
  /* iteration of sMapEnclosedBy() until FALSE 
     returns TRUE if first call to sMapEnclosedBy returns TRUE
  */

int sMapLength (KEY key) ;
/* try various strategies (DNA, S_parent, S_children) to find the sequence 
   length of an object
*/

#endif /* DEF_SMAP_H */
/***************************************************************************/
